# --------------------------
# -------- OVERVIEW --------
# --------------------------
#
# TCL script written by Ben Carbery (ben.carbery@anu.edu.au).
#
# Cacti script for retrieving and parsing a bind statistics file
# - Assumes the stats file is being generated and replaced every 
#   cacti polling cycle (not appended to)
# - Stats file can be generated by 'rndc stats' in a cron job on
#   bind server
# - This script outputs over 50 data items. This can cause errors
#   in cacti poller if the total characters in output exceed 1121
#   The solution is to implement results sets. Choose one of 'in',
#   'out' or 'misc' to select appropriate output for your graph
#   using the -r option

# ---------------------------
# -------- CHANGELOG --------
# ---------------------------
#
# 20110317 (BC)
#     Initial version

# ----------------------------------------
# -------- CAVEATS / BUGS / TO DO --------
# ----------------------------------------

# -----------------------------
# -------- GLOBAL VARS --------
# -----------------------------

proc listifyFile {fileName {outputListName fileContents} {nocomments 1} {noblanks 0}} {
    upvar 1 $outputListName outputList
    set outputList [list]

    if { [catch {set fd [open $fileName r] } catchError] } {
        puts stderr "ERROR: $catchError"
        return 0
    } else {
        foreach line [split [read -nonewline $fd] \n] {
            # Trim carriage returns
            set line [string trim $line \r]

            # Filter comments?
            if { $nocomments } {
                if { [regexp {^#} $line] } {
                    continue
                }
            }

            # Filter blank lines?
            if { $noblanks } {
                if { [regexp {^$} $line] } {
                    continue
                }
            }

            # Line passed filters
            lappend outputList $line
        }
        close $fd
        return 1
    }
}

# -----------------------------
# -------- GLOBAL VARS --------
# -----------------------------

set usage "Usage: $argv0 -h <hostname> -u <scp_user> <?-r results_set?> <?-s stats_file?>"


set scpBinary "/usr/bin/scp"
set scpOpts "-p"

set resultsSet "in"
set statsFile "/var/namedb/stats/named.stats"
set statsFileName [file tail $statsFile]
set dstDir "/var/www/cacti/scripts"

# ------------------------------
# -------- COMMAND LINE --------
# ------------------------------

if { [llength $argv] == 0 } {
    puts stderr $usage
    exit 1
}

while { [string match -* [set option [lindex $argv 0]]] == 1 } {

    # Find the matching option.
    # Note: the "--" below only marks the end of options to switch itself.
    switch -- $option {
        -h {
            set hostname [lindex $argv 1]
            set argv [lreplace $argv 0 1]
        }
        -r {
            set resultsSet [lindex $argv 1]
            set argv [lreplace $argv 0 1]
        }
        -s {
            set statsFile [lindex $argv 1]
            set statsFileName [file tail $statsFile]
            set argv [lreplace $argv 0 1]
        }
        -u {
            set user [lindex $argv 1]
            set argv [lreplace $argv 0 1]
        }
        default {
            puts stderr $usage
            exit 1
        }
    }
}

# --------------------------
# -------- GET FILE --------
# --------------------------

if { [catch {exec $scpBinary $scpOpts $user@$hostname:$statsFile $dstDir/$statsFileName.$hostname} catchError] } {
    puts stderr "ERROR: Could not retrieve $statsFile from $hostname: $catchError"
    exit 1
}

# ----------------------------
# -------- PARSE FILE --------
# ----------------------------

if { ![listifyFile "$dstDir/$statsFileName.$hostname" fileContents 0 0] } {
    exit 1
}

# Results variables
set resultsGeneral [list]
set resultsIn [list]
set resultsOut [list]
set resultsMisc [list]
set section 0

foreach line $fileContents {
    # Check we've reached end of a dump section
    # The script normally expect there to be only 1 section, but named will by default append to a stats file
    if { [regexp {^--- Statistics Dump ---} $line] } {
        break

    } elseif { [string equal "++ Incoming Requests ++" $line] } {
        set section 1

    } elseif { [string equal "++ Incoming Queries ++" $line] } {
        set section 2

    } elseif { [string equal "++ Outgoing Queries ++" $line] } {
        set section 3

    } elseif { [string equal "++ Name Server Statistics ++" $line] } {
        set section 4

    } elseif { [string equal "++ Zone Maintenance Statistics ++" $line] } {
        set section 5

    } elseif { [string equal "++ Resolver Statistics ++" $line] } {
        set section 6

    } elseif { [string equal "++ Cache DB RRsets ++" $line] } {
        set section 7

    } elseif { [string equal "++ Socket I/O Statistics ++" $line] } {
        set section 8

    } elseif { [string equal "++ Per Zone Query Statistics ++" $line] } {
        set section 9

    } elseif { [regexp {^\+\+.+\+\+$} $line] } {
        set section 0
    
    # Incoming Requests
    } elseif { $section == 1 } {
        if { [regexp {^\s+(\d+)\s+QUERY$} $line xxx total_queries] } {
            lappend resultsGeneral "TOTAL_QUERIES:$total_queries"

        } elseif { [regexp {^\s+(\d+)\s+NOTIFY$} $line xxx total_notifies] } {
            lappend resultsGeneral "TOTAL_NOTIFIES:$total_notifies"

        } elseif { [regexp {^\s+(\d+)\s+UPDATE$} $line xxx total_updates] } {
            lappend resultsGeneral "TOTAL_UPDATES:$total_updates"

        }

    # Incoming Queries
    } elseif { $section == 2 } {
        if { [regexp {^\s+(\d+)\s+A$} $line xxx queries_in_a] } {
            lappend resultsIn "QUERIES_IN_A:$queries_in_a"

        } elseif { [regexp {^\s+(\d+)\s+A6$} $line xxx queries_in_a6] } {
            lappend resultsIn "QUERIES_IN_A6:$queries_in_a6"

        } elseif { [regexp {^\s+(\d+)\s+AAAA$} $line xxx queries_in_aaaa] } {
            lappend resultsIn "QUERIES_IN_AAAA:$queries_in_aaaa"

        } elseif { [regexp {^\s+(\d+)\s+AFSDB$} $line xxx queries_in_afsdb] } {
            lappend resultsIn "QUERIES_IN_AFSDB:$queries_in_afsdb"

        } elseif { [regexp {^\s+(\d+)\s+ANY$} $line xxx queries_in_any] } {
            lappend resultsIn "QUERIES_IN_ANY:$queries_in_any"

        } elseif { [regexp {^\s+(\d+)\s+AXFR$} $line xxx queries_in_axfr] } {
            lappend resultsIn "QUERIES_IN_AXFR:$queries_in_axfr"

        } elseif { [regexp {^\s+(\d+)\s+CNAME$} $line xxx queries_in_cname] } {
            lappend resultsIn "QUERIES_IN_CNAME:$queries_in_cname"

        } elseif { [regexp {^\s+(\d+)\s+DNSKEY$} $line xxx queries_in_dnskey] } {
            lappend resultsIn "QUERIES_IN_DNSKEY:$queries_in_dnskey"

        } elseif { [regexp {^\s+(\d+)\s+DS$} $line xxx queries_in_ds] } {
            lappend resultsIn "QUERIES_IN_DS:$queries_in_ds"

        } elseif { [regexp {^\s+(\d+)\s+HINFO$} $line xxx queries_in_hinfo] } {
            lappend resultsIn "QUERIES_IN_HINFO:$queries_in_hinfo"

        } elseif { [regexp {^\s+(\d+)\s+IXFR$} $line xxx queries_in_ixfr] } {
            lappend resultsIn "QUERIES_IN_IXFR:$queries_in_ixfr"

        } elseif { [regexp {^\s+(\d+)\s+MAILB$} $line xxx queries_in_mailb] } {
            lappend resultsIn "QUERIES_IN_MAILB:$queries_in_mailb"

        } elseif { [regexp {^\s+(\d+)\s+MX$} $line xxx queries_in_mx] } {
            lappend resultsIn "QUERIES_IN_MX:$queries_in_mx"

        } elseif { [regexp {^\s+(\d+)\s+NAPTR$} $line xxx queries_in_naptr] } {
            lappend resultsIn "QUERIES_IN_NAPTR:$queries_in_naptr"

        } elseif { [regexp {^\s+(\d+)\s+NS$} $line xxx queries_in_ns] } {
            lappend resultsIn "QUERIES_IN_NS:$queries_in_ns"

        } elseif { [regexp {^\s+(\d+)\s+NSEC$} $line xxx queries_in_nsec] } {
            lappend resultsIn "QUERIES_IN_NSEC:$queries_in_nsec"

        } elseif { [regexp {^\s+(\d+)\s+Others$} $line xxx queries_in_others] } {
            lappend resultsIn "QUERIES_IN_OTHERS:$queries_in_others"

        } elseif { [regexp {^\s+(\d+)\s+PTR$} $line xxx queries_in_ptr] } {
            lappend resultsIn "QUERIES_IN_PTR:$queries_in_ptr"

        } elseif { [regexp {^\s+(\d+)\s+RESERVED0$} $line xxx queries_in_reserved0] } {
            lappend resultsIn "QUERIES_IN_RESERVED0:$queries_in_reserved0"

        } elseif { [regexp {^\s+(\d+)\s+RRSIG$} $line xxx queries_in_rrsig] } {
            lappend resultsIn "QUERIES_IN_RRSIG:$queries_in_rrsig"

        } elseif { [regexp {^\s+(\d+)\s+SOA$} $line xxx queries_in_soa] } {
            lappend resultsIn "QUERIES_IN_SOA:$queries_in_soa"

        } elseif { [regexp {^\s+(\d+)\s+SRV$} $line xxx queries_in_srv] } {
            lappend resultsIn "QUERIES_IN_SRV:$queries_in_srv"

        } elseif { [regexp {^\s+(\d+)\s+SSHFP$} $line xxx queries_in_sshfp] } {
            lappend resultsIn "QUERIES_IN_SSHFP:$queries_in_sshfp"

        } elseif { [regexp {^\s+(\d+)\s+TKEY$} $line xxx queries_in_tkey] } {
            lappend resultsIn "QUERIES_IN_TKEY:$queries_in_tkey"

        } elseif { [regexp {^\s+(\d+)\s+TXT$} $line xxx queries_in_txt] } {
            lappend resultsIn "QUERIES_IN_TXT:$queries_in_txt"

        } elseif { [regexp {^\s+(\d+)\s+TYPE55$} $line xxx queries_in_type55] } {
            lappend resultsIn "QUERIES_IN_TYPE55:$queries_in_type55"

        } elseif { [regexp {^\s+(\d+)\s+X25$} $line xxx queries_in_x25] } {
            lappend resultsIn "QUERIES_IN_X25:$queries_in_x25"

        }

    # Outgoing Queries
    } elseif { $section == 3 } {
        if { [regexp {^\s+(\d+)\s+A$} $line xxx queries_out_a] } {
            lappend resultsOut "QUERIES_OUT_A:$queries_out_a"

        } elseif { [regexp {^\s+(\d+)\s+A6$} $line xxx queries_out_a6] } {
            lappend resultsOut "QUERIES_OUT_A6:$queries_out_a6"

        } elseif { [regexp {^\s+(\d+)\s+AFSDB$} $line xxx queries_out_afsdb] } {
            lappend resultsOut "QUERIES_OUT_AFSDB:$queries_out_afsdb"

        } elseif { [regexp {^\s+(\d+)\s+AAAA$} $line xxx queries_out_aaaa] } {
            lappend resultsOut "QUERIES_OUT_AAAA:$queries_out_aaaa"

        } elseif { [regexp {^\s+(\d+)\s+ANY$} $line xxx queries_out_any] } {
            lappend resultsOut "QUERIES_OUT_ANY:$queries_out_any"

        } elseif { [regexp {^\s+(\d+)\s+CNAME$} $line xxx queries_out_cname] } {
            lappend resultsOut "QUERIES_OUT_CNAME:$queries_out_cname"

        } elseif { [regexp {^\s+(\d+)\s+HINFO$} $line xxx queries_out_hinfo] } {
            lappend resultsOut "QUERIES_OUT_HINFO:$queries_out_hinfo"

        } elseif { [regexp {^\s+(\d+)\s+MX$} $line xxx queries_out_mx] } {
            lappend resultsOut "QUERIES_OUT_MX:$queries_out_mx"

        } elseif { [regexp {^\s+(\d+)\s+NAPTR$} $line xxx queries_out_naptr] } {
            lappend resultsOut "QUERIES_OUT_NAPTR:$queries_out_naptr"

        } elseif { [regexp {^\s+(\d+)\s+Others$} $line xxx queries_out_others] } {
            lappend resultsOut "QUERIES_OUT_OTHERS:$queries_out_others"

        } elseif { [regexp {^\s+(\d+)\s+PTR$} $line xxx queries_out_ptr] } {
            lappend resultsOut "QUERIES_OUT_PTR:$queries_out_ptr"

        } elseif { [regexp {^\s+(\d+)\s+SOA$} $line xxx queries_out_soa] } {
            lappend resultsOut "QUERIES_OUT_SOA:$queries_out_soa"

        } elseif { [regexp {^\s+(\d+)\s+SPF$} $line xxx queries_out_spf] } {
            lappend resultsOut "QUERIES_OUT_SPF:$queries_out_spf"

        } elseif { [regexp {^\s+(\d+)\s+SRV$} $line xxx queries_out_srv] } {
            lappend resultsOut "QUERIES_OUT_SRV:$queries_out_srv"

        } elseif { [regexp {^\s+(\d+)\s+SSHFP$} $line xxx queries_out_sshfp] } {
            lappend resultsOut "QUERIES_OUT_SSHFP:$queries_out_sshfp"

        } elseif { [regexp {^\s+(\d+)\s+TXT$} $line xxx queries_out_txt] } {
            lappend resultsOut "QUERIES_OUT_TXT:$queries_out_txt"

        }

    # Name Server Statistics
    } elseif { $section == 4 } {
        if { [regexp {^\s+(\d+)\s+IPv4 requests received$} $line xxx requests_rcvd_ipv4] } {
            lappend resultsMisc "REQUESTS_RCVD_IPV4:$requests_rcvd_ipv4"

        } elseif { [regexp {^\s+(\d+)\s+responses sent$} $line xxx responses_sent] } {
            lappend resultsMisc "RESPONSES_SENT:$responses_sent"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in successful answer$} $line xxx success_answered] } {
            lappend resultsMisc "ANSWERED_SUCCESS:$success_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in authoritative answer$} $line xxx auth_answered] } {
            lappend resultsMisc "ANSWERED_AUTH:$auth_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in non authoritative answer$} $line xxx nonauth_answered] } {
            lappend resultsMisc "ANSWERED_NONAUTH:$nonauth_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in referral answer$} $line xxx referral_answered] } {
            lappend resultsMisc "ANSWERED_REFERRAL:$referral_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in nxrrset$} $line xxx nxrrset_answered] } {
            lappend resultsMisc "ANSWERED_NXRRSET:$nxrrset_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in SERVFAIL$} $line xxx servfail_answered] } {
            lappend resultsMisc "ANSWERED_SERVFAIL:$servfail_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries resulted in NXDOMAIN$} $line xxx nxdomain_answered] } {
            lappend resultsMisc "ANSWERED_NXDOMAIN:$nxdomain_answered"

        } elseif { [regexp {^\s+(\d+)\s+queries caused recursion$} $line xxx caused_recursion] } {
            lappend resultsMisc "CAUSED_RECURSION:$caused_recursion"

        } elseif { [regexp {^\s+(\d+)\s+duplicate queries received$} $line xxx duplicates] } {
            lappend resultsMisc "DUPLICATE_QUERIES:$duplicates"

        } elseif { [regexp {^\s+(\d+)\s+queries dropped$} $line xxx dropped] } {
            lappend resultsMisc "DROPPED_QUERIES:$dropped"

        } elseif { [regexp {^\s+(\d+)\s+other query failures$} $line xxx other_failures] } {
            lappend resultsMisc "OTHER_FAILURES:$other_failures"

        }

    # Zone Maintenance Statistics
    } elseif { $section == 5 } {
        continue

    # Resolver Statistics
    } elseif { $section == 6 } {
        if { [regexp {^\s+(\d+)\s+IPv4 queries sent$} $line xxx queries_sent_ipv4] } {
            lappend resultsMisc "QUERIES_OUT_IPV4:$queries_sent_ipv4"

        } elseif { [regexp {^\s+(\d+)\s+IPv6 queries sent$} $line xxx queries_sent_ipv6] } {
            lappend resultsMisc "QUERIES_OUT_IPV6:$queries_sent_ipv6"

        } elseif { [regexp {^\s+(\d+)\s+IPv4 responses received$} $line xxx responses_rcvd_ipv4] } {
            lappend resultsMisc "RESPONSES_IN_IPV4:$responses_rcvd_ipv4"

        } elseif { [regexp {^\s+(\d+)\s+IPv6 responses received$} $line xxx responses_rcvd_ipv6] } {
            lappend resultsMisc "RESPONSES_IN_IPV6:$responses_rcvd_ipv6"

        }

    # Cache DB RRsets
    } elseif { $section == 7 } {
        continue

    # Socket I/O Statistics
    } elseif { $section == 8 } {
        if { [regexp {^\s+(\d+)\s+UDP/IPv4 connections established$} $line xxx conns_est_ipv4] } {
            lappend resultsMisc "CONNS_EST_IPV4:$conns_est_ipv4"

        } elseif { [regexp {^\s+(\d+)\s+UDP/IPv6 connections established$} $line xxx conns_est_ipv6] } {
            lappend resultsMisc "CONNS_EST_IPV6:$conns_est_ipv6"

        }

    # Per Zone Query Statistics
    } elseif { $section == 9 } {
        continue

    }
}

# Prep Results
switch -- $resultsSet {
    in {
        set resultsOut [list]
        set resultsMisc [list]
    }
    out {
        set resultsIn [list]
        set resultsMisc [list]
    }
    misc {
        set resultsIn [list]
        set resultsOut [list]
    }
    default {
        puts stderr "Invalid results set requested: $resultsSet"
        exit 1
    }
}

# Send results to stdout
# set results [string trim $results]
puts [string trim [concat $resultsGeneral $resultsIn $resultsOut $resultsMisc]]
